//
//  SFRecordable.m
//  Sketch
//
//  Created by doster on 12/15/05.
//  Copyright 2005 Apple Computer, Inc. All rights reserved.
// As of OS X 10.4 (Tiger), AppKit's scripting support only recognizes object references as direct objects of
// AppleScript commands. In particular, the following aren't supported:
// aeLists of object referencs,  "thru" clauses, and whose clauses.
//
// The API here was designed to be cheap when recording was off.
// Some modifications by David Phillip Oster. You may use those modifications freely, provided that just
// as you hold Apple blameless, you must also hold me blameless. Feb 2006 <mailto:oster@ieee.org>


#import "SFRecordable.h"
#import	"SKTGraphic.h"

static SFRecordingManager* sSharedInstance = nil;

@interface NSAppleEventDescriptor (SFREcordingExtensions )
+ (NSAppleEventDescriptor*) descriptorWithFloat: (float) aFloat;
@end

@interface NSColor (SFREcordingExtensions )
- (NSAppleEventDescriptor*) aeDescriptor;
@end

@interface NSMutableAttributedString (SFREcordingExtensions )
- (NSAppleEventDescriptor*) aeDescriptor;
@end

@interface NSScriptObjectSpecifier (AEDescriptorExtension)
- (NSAppleEventDescriptor*) aeDescriptor;
@end

@interface NSString (FourCharCodeExtension)
- (FourCharCode) fourCharCode;
@end

@implementation NSString (FourCharCodeExtension)
// Surely this must already be in AppKit, but I couldn't find it.
- (FourCharCode) fourCharCode {
	FourCharCode val = 0;
	NSAssert(4 == [self length], @"must be 4 long");
	memcpy(&val, [self cStringUsingEncoding: NSMacOSRomanStringEncoding], sizeof val);
	return val;
}
@end


@implementation NSScriptObjectSpecifier (AEDescriptorExtension)
// I've filed a Radar to make _asDescriptor a public SPI
- (NSAppleEventDescriptor*) aeDescriptor {
	return [self performSelector: @selector(_asDescriptor)];
}
@end

@implementation NSMutableAttributedString (SFREcordingExtensions )
// Doesn't do it justice: looses the attributes
- (NSAppleEventDescriptor*) aeDescriptor {
	return [NSAppleEventDescriptor descriptorWithString: [self string]];
}
@end


@implementation SFRecordingManager

// Creates a AppleEvent for recording, and sets the direct object specifier. OK to pass "nil" for direct object. Caller must release!
- (NSAppleEventDescriptor*) createRecordingEvent: (NSString*) commmand directSpecifier: (NSScriptObjectSpecifier*) spec{
	NSScriptObjectSpecifier *containerRef = [NSApp  objectSpecifier];
	ProcessSerialNumber selfPsn = {0, kCurrentProcess};
	NSAppleEventDescriptor* aeAddress = [NSAppleEventDescriptor descriptorWithDescriptorType: typeProcessSerialNumber bytes: &selfPsn length: sizeof selfPsn];
	AEEventClass aeClass =  [[commmand substringWithRange: NSMakeRange(0, 4)] fourCharCode];
	AEEventID aeEvent =		[[commmand substringWithRange: NSMakeRange(4, 4)] fourCharCode];
	NSAppleEventDescriptor* ae = 
			[[NSAppleEventDescriptor alloc] initWithEventClass: aeClass 
								eventID: aeEvent 
					   targetDescriptor: aeAddress 
							   returnID: kAutoGenerateReturnID 
						  transactionID: kAnyTransactionID];
	if( nil != spec){
		[ae setDescriptor: [spec aeDescriptor] forKeyword: keyDirectObject];
	}
	return ae;
}

// Actually sends the recording event. Separated from "create" so additional setDescriptor: calls can be done.
- (void) sendRecordingEvent: (NSAppleEventDescriptor*) ae {
	if( nil != ae){
		AEDesc reply = {typeNull, nil};
		AESend([ae aeDesc], &reply, kAEDontExecute | kAENoReply, kAENormalPriority, 0, nil, nil);
		AEDisposeDesc(&reply);
	}
}


- (void) recordCommand: (NSString*) commmand directSpecifier: (NSScriptObjectSpecifier*) spec aeValue: (NSAppleEventDescriptor*) value key: (NSString*) key {
	NSAppleEventDescriptor* event = [self createRecordingEvent: commmand directSpecifier: spec];
	[event setDescriptor: value forKeyword: [key fourCharCode]];
	[self sendRecordingEvent: event];
	[event release];
}


// Handle recording a simple command that just needs an 8-byte specifier and a direct object. if direct object is nil, it is omitted from event.
- (void) recordCommand: (NSString*) commmand directSpecifier: (NSScriptObjectSpecifier*) spec {
	NSAppleEventDescriptor* event = [self createRecordingEvent: commmand directSpecifier: spec];
	[self sendRecordingEvent: event];
	[event release];
}

// 
- (NSScriptObjectSpecifier*) specifierFromDirectObject: (id) directObj {
	NSScriptObjectSpecifier* val = nil;
	if( mIsRecording && [directObj respondsToSelector: @selector(objectSpecifier) ]){
		val = [directObj performSelector: @selector(objectSpecifier)];
	}
	return val;
}


// >>> not yet: 
- (void) recordIndex1: (id) directObj {
	NSScriptObjectSpecifier* directSpecifier = nil;
	if( nil != (directSpecifier = [self specifierFromDirectObject: directObj])){
	}
}


- (void) recordSet1: (id) directObj property: (NSString*) propertyName to: (NSAppleEventDescriptor*) value {
	NSScriptObjectSpecifier* directSpecifier = nil;
	if( nil != (directSpecifier = [self specifierFromDirectObject: directObj])){
		NSScriptObjectSpecifier* propSpec = [[NSPropertySpecifier alloc] initWithContainerSpecifier: directSpecifier key: propertyName];
		[self recordCommand: @"coresetd" directSpecifier: propSpec aeValue: value key: @"data"];
		[propSpec release];
	}
}

- (void) recordPosition1: (id) directObj {
	[self recordSet1: directObj property: @"xPosition" to: [NSAppleEventDescriptor descriptorWithFloat: [(SKTGraphic*) directObj xPosition]]];
	[self recordSet1: directObj property: @"yPosition" to: [NSAppleEventDescriptor descriptorWithFloat: [(SKTGraphic*) directObj yPosition]]];
}

- (void) recordSize1: (id) directObj {
	[self recordSet1: directObj property: @"width" to: [NSAppleEventDescriptor descriptorWithFloat: [(SKTGraphic*) directObj width]]];
	[self recordSet1: directObj property: @"height" to: [NSAppleEventDescriptor descriptorWithFloat: [(SKTGraphic*) directObj height]]];
}


- (void) recordBounds1: (id) directObj {
	NSScriptObjectSpecifier* directSpecifier = nil;
	if( nil != (directSpecifier = [self specifierFromDirectObject: directObj])){
		[self recordPosition1: directObj];
		[self recordSize1: directObj];
	}
}
- (void) recordDelete1: (id) directObj {
	NSScriptObjectSpecifier* directSpecifier = nil;
	if( nil != (directSpecifier = [self specifierFromDirectObject: directObj])){
		[self recordCommand: @"coredelo" directSpecifier: directSpecifier];
	}
}


- (void) forEach: (id) directObj perform: (SEL) selectr {
	int i, iLen;

	if( !  mIsRecording){
		/* if not recording, do nothing */
	}else if( [directObj respondsToSelector: @selector(count)] && 
		0 < (iLen = (int) [directObj performSelector: @selector(count)]) && 
		[directObj respondsToSelector: @selector(objectAtIndex:)]){

		for(i = 0;i < iLen; ++i){
			[self performSelector: selectr withObject: [directObj objectAtIndex: i]];
		}
	}else{
		[self performSelector: selectr withObject: directObj];
	}
}

- (void) handleRecordingOnAppleEvent: (NSAppleEventDescriptor*) event withReplyEvent: (NSAppleEventDescriptor*) replyEvent
{
	mIsRecording = YES;
}

- (void) handleRecordingOffAppleEvent: (NSAppleEventDescriptor*) event withReplyEvent: (NSAppleEventDescriptor*) replyEvent
{
	mIsRecording = NO;
}

// ====================== public API ==============
#pragma mark public API

+ (SFRecordingManager*) sharedInstance {
	if( nil == sSharedInstance){
		sSharedInstance = [[SFRecordingManager alloc] init];
	}
	return sSharedInstance;
}

// we assume that recording is on initially, because we don't get our handleRecordingOnAppleEvent: installed early enough to get the info from the OS.
- (id) init {
	NSAssert(nil == sSharedInstance, @"There can be only one.");
	if( nil != (self = [super init])){
		mIsRecording = YES;
	}
	return self;
}


// Apps should call this from applicationDidFinishLaunching to
// installs handlers to listen for notifications that recording is going on/off.
- (void) finishLaunching
{
	NSAppleEventManager *appleEventManager = [NSAppleEventManager sharedAppleEventManager];
	[appleEventManager setEventHandler: self andSelector: @selector(handleRecordingOnAppleEvent: withReplyEvent:) forEventClass: kCoreEventClass andEventID: kAENotifyStartRecording];
	[appleEventManager setEventHandler: self andSelector: @selector(handleRecordingOffAppleEvent: withReplyEvent:) forEventClass: kCoreEventClass andEventID: kAENotifyStopRecording];
}

// read-only: clients can query this to see if recording is on.
- (BOOL) isRecording
{
	return mIsRecording;
}

// We pass a raw (id) here, so we can only convert it to a NSAppleEventDescriptor if we are recording. 
- (void) recordSet: (id) directObj property: (NSString*) propertyName to: (id) rawValue {
	// if not recording, do nothing
	if( mIsRecording){

		NSAppleEventDescriptor* value = nil;
		if( [rawValue respondsToSelector: @selector(aeDescriptor)]){
			value = [rawValue aeDescriptor];
		}else if( [rawValue respondsToSelector: @selector(objCType)]){
			const char* cType = [rawValue objCType];
			if( 0 != cType[0] && 0 == cType[1]){
				switch(cType[0]){
				case 'c': value = [NSAppleEventDescriptor descriptorWithBoolean: [rawValue boolValue]];	break;
				case 'i': value = [NSAppleEventDescriptor descriptorWithInt32: [rawValue intValue]];	break;
				case 'f': value = [NSAppleEventDescriptor descriptorWithFloat: [rawValue floatValue]];	break;
				}
			}
		}

		if( nil != value){
			int i, iLen;
			if( [directObj respondsToSelector: @selector(count)] && 
				0 < (iLen = (int) [directObj performSelector: @selector(count)]) && 
				[directObj respondsToSelector: @selector(objectAtIndex:)]){

				for(i = 0;i < iLen; ++i){
					[self recordSet1: [directObj objectAtIndex: i] property: propertyName to: value];
				}
			}else{
				[self recordSet1: directObj property: propertyName to: value];
			}
		}
	}
}

// it would be nice if we could query to find out what the important properties are.
//
// Note:
// http://developer.apple.com/documentation/AppleScript/Conceptual/AppleEvents/debugging_aepg/chapter_9_section_4.html
// I used the following shell commands to help debug this:
//$ export AEDebugSends=1; export AEDebugReceives=1
//$ osascript -e "tell application \"Sketch\" to get properties of graphic 1 of document \"Untitled\""

- (void) recordCreate: (id) createdObj {
	NSScriptObjectSpecifier* directSpecifier = nil;
	if( nil != (directSpecifier = [self specifierFromDirectObject: createdObj])){
		// Note: we've gone through the classes of the model and added methods to return the class code. A better design would look up the ObjC class in the NSScriptSuiteRegistry
		NSAppleEventDescriptor* classAE = [NSAppleEventDescriptor descriptorWithTypeCode: [createdObj appleEventCode]];

		// Note: assumes we are creating at beginning
		NSAppleEventDescriptor* insertionLocationAE = [NSAppleEventDescriptor recordDescriptor];
		[insertionLocationAE setParamDescriptor: [[directSpecifier containerSpecifier] aeDescriptor] forKeyword: keyAEObject];
		[insertionLocationAE setParamDescriptor: [NSAppleEventDescriptor descriptorWithEnumCode: kAEBeginning] forKeyword: keyAEPosition];

		// Note: assumes important properties are the same for all objects
		NSAppleEventDescriptor* propertiesAE = [NSAppleEventDescriptor recordDescriptor];
		[propertiesAE setParamDescriptor: [NSAppleEventDescriptor descriptorWithFloat: [createdObj xPosition]] forKeyword: 'xpos'];
		[propertiesAE setParamDescriptor: [NSAppleEventDescriptor descriptorWithFloat: [createdObj yPosition]] forKeyword: 'ypos'];
		[propertiesAE setParamDescriptor: [NSAppleEventDescriptor descriptorWithFloat: [createdObj width]] forKeyword: 'widt'];
		[propertiesAE setParamDescriptor: [NSAppleEventDescriptor descriptorWithFloat: [createdObj height]] forKeyword: 'heig'];
		if( [createdObj canDrawStroke]){
			[propertiesAE setParamDescriptor: [NSAppleEventDescriptor descriptorWithFloat: [createdObj strokeLineWidth]] forKeyword: 'slwd'];
			[propertiesAE setParamDescriptor: [NSAppleEventDescriptor descriptorWithFloat: [createdObj strokeAlpha]] forKeyword: 'salp'];
			[propertiesAE setParamDescriptor: [[[createdObj strokeColor] colorUsingColorSpaceName: NSDeviceRGBColorSpace] aeDescriptor] forKeyword: 'sclr'];
		}
		if( [createdObj canDrawFill]){
			[propertiesAE setParamDescriptor: [NSAppleEventDescriptor descriptorWithFloat: [createdObj fillAlpha]] forKeyword: 'falp'];
			[propertiesAE setParamDescriptor: [[[createdObj fillColor] colorUsingColorSpaceName: NSDeviceRGBColorSpace] aeDescriptor] forKeyword: 'fclr'];
		}

		NSAppleEventDescriptor* event = [self createRecordingEvent: @"corecrel" directSpecifier: nil];
		[event setDescriptor: classAE forKeyword: keyAEObjectClass];
		[event setDescriptor: [insertionLocationAE coerceToDescriptorType: typeInsertionLoc] forKeyword: keyAEInsertHere];
		[event setDescriptor: propertiesAE forKeyword: keyAEPropData];
		[self sendRecordingEvent: event];
		[event release];
	}
}

// >>> not yet: we need both before and after information here.
- (void) recordIndex: (id) directObj {
	[self forEach: directObj perform: @selector(recordIndex1:)];
}

- (void) recordDelete: (id) directObj {
	[self forEach: directObj perform: @selector(recordDelete1:)];
}

- (void) recordBounds: (id) directObj {
	[self forEach: directObj perform: @selector(recordBounds1:)];
}

- (void) recordPosition: (id) directObj {
	[self forEach: directObj perform: @selector(recordPosition1:)];
}

- (void) recordSize: (id) directObj {
	[self forEach: directObj perform: @selector(recordSize1:)];
}

@end



@implementation NSAppleEventDescriptor (SFREcordingExtensions )
+ (NSAppleEventDescriptor*) descriptorWithFloat: (float) aFloat {
	return [NSAppleEventDescriptor descriptorWithDescriptorType: typeIEEE32BitFloatingPoint bytes: &aFloat length: sizeof aFloat];
}

@end

@implementation NSColor (SFREcordingExtensions )
- (NSAppleEventDescriptor*) aeDescriptor {
	float red, green, blue;
	RGBColor rgb;
	[self  getRed: &red green: &green blue: &blue alpha: nil];
	rgb.red = 0xFFFFU * red;
	rgb.green = 0xFFFFU * green;
	rgb.blue = 0xFFFFU * blue;
	return [NSAppleEventDescriptor descriptorWithDescriptorType: typeRGBColor bytes: &rgb length: sizeof rgb];
}
@end


