// SKTGraphicView.m
// Sketch Example
//

#import "SKTGraphicView.h"
#import "SKTDrawWindowController.h"
#import "SKTDrawDocument.h"
#import "SKTGraphic.h"
#import "SKTFoundationExtras.h"
#import "SKTToolPaletteController.h"
#import "SKTImage.h"
#import "SKTGridPanelController.h"

NSString *SKTGraphicViewSelectionDidChangeNotification = @"SKTGraphicViewSelectionDidChange";

@implementation SKTGraphicView

static float SKTDefaultPasteCascadeDelta = 10.0;

- (id) initWithFrame: (NSRect) frame {
    self = [super initWithFrame: frame];
    if( self) {
        NSMutableArray *dragTypes = [NSMutableArray arrayWithObjects: NSColorPboardType, NSFilenamesPboardType, nil];
        [dragTypes addObjectsFromArray: [NSImage imagePasteboardTypes]];
        [self registerForDraggedTypes: dragTypes];
        selectedGraphics_ = [[NSMutableArray allocWithZone: [self zone]] init];
        creatingGraphic_ = nil;
        rubberbandRect_ = NSZeroRect;
        rubberbandGraphics_ = nil;
        gvFlags_.rubberbandIsDeselecting = NO;
        gvFlags_.initedRulers = NO;
        editingGraphic_ = nil;
        editorView_ = nil;
        pasteboardChangeCount_ = -1;
        pasteCascadeNumber_ = 0;
        pasteCascadeDelta_ = NSMakePoint(SKTDefaultPasteCascadeDelta, SKTDefaultPasteCascadeDelta);
        gvFlags_.snapsToGrid = NO;
        gvFlags_.showsGrid = NO;
        gvFlags_.knobsHidden = NO;
        gridSpacing_ = 8.0;
        gridColor_ = [[NSColor lightGrayColor] retain];
        unhideKnobsTimer_ = nil;
    }
    return self;
}

- (void) dealloc {
    [self endEditing];
    [selectedGraphics_ release];
    [rubberbandGraphics_ release];
    [gridColor_ release];
    [super dealloc];
}

// SKTDrawWindowController accessors and convenience methods
- (void) setDrawWindowController: (SKTDrawWindowController*) theController {
    controller = theController;
}

- (SKTDrawWindowController*) drawWindowController {
    return controller;
}

- (SKTDrawDocument*) drawDocument {
    return [[self drawWindowController] document];
}

- (NSArray*) graphics {
    return [[[self drawWindowController] document] graphics];
}

// Display invalidation
- (void) invalidateGraphic: (SKTGraphic*) graphic {
    [self setNeedsDisplayInRect: [graphic drawingBounds]];
    if( ![[self graphics] containsObject: graphic]) {
        [self deselectGraphic: graphic];  // deselectGraphic will call invalidateGraphic, too, but only if the graphic is in the selection and since the graphic is removed from the selection before this method is called again the potential infinite loop should not happen.
    }
}

- (void) invalidateGraphics: (NSArray*) graphics {
    unsigned i, c = [graphics count];
    for (i=0; i<c; i++) {
        [self invalidateGraphic: [graphics objectAtIndex: i]];
    }
}

// Selection primitives
- (NSArray*) selectedGraphics {
    return selectedGraphics_;
}

static int SKT_orderGraphicsFrontToBack(id graphic1, id graphic2, void *gArray) {
    NSArray *graphics = (NSArray*) gArray;
    unsigned index1, index2;

    index1 = [graphics indexOfObjectIdenticalTo: graphic1];
    index2 = [graphics indexOfObjectIdenticalTo: graphic2];
    if( index1 == index2) {
        return NSOrderedSame;
    } else if( index1 < index2) {
        return NSOrderedAscending;
    } else {
        return NSOrderedDescending;
    }
}

- (NSArray*) orderedSelectedGraphics  {
    return [[self selectedGraphics] sortedArrayUsingFunction: SKT_orderGraphicsFrontToBack context: [self graphics]];
}

- (BOOL) graphicIsSelected: (SKTGraphic*) graphic {
    return (([selectedGraphics_ indexOfObjectIdenticalTo: graphic] == NSNotFound) ? NO : YES);
}

- (void) selectGraphic: (SKTGraphic*) graphic {
    unsigned curIndex = [selectedGraphics_ indexOfObjectIdenticalTo: graphic];
    if( curIndex == NSNotFound) {
        [[[self undoManager] prepareWithInvocationTarget: self] deselectGraphic: graphic];
        [[self undoManager] setActionName: NSLocalizedStringFromTable(@"Selection Change", @"UndoStrings", @"Action name for selection changes.")];
        [selectedGraphics_ addObject: graphic];
        [self invalidateGraphic: graphic];
        pasteCascadeDelta_ = NSMakePoint(SKTDefaultPasteCascadeDelta, SKTDefaultPasteCascadeDelta);
        [[NSNotificationCenter defaultCenter] postNotificationName: SKTGraphicViewSelectionDidChangeNotification object: self];
        [self updateRulers];
    }
}

- (void) deselectGraphic: (SKTGraphic*) graphic {
    unsigned curIndex = [selectedGraphics_ indexOfObjectIdenticalTo: graphic];
    if( curIndex != NSNotFound) {
        [[[self undoManager] prepareWithInvocationTarget: self] selectGraphic: graphic];
        [[self undoManager] setActionName: NSLocalizedStringFromTable(@"Selection Change", @"UndoStrings", @"Action name for selection changes.")];
        [selectedGraphics_ removeObjectAtIndex: curIndex];
        [self invalidateGraphic: graphic];
        pasteCascadeDelta_ = NSMakePoint(SKTDefaultPasteCascadeDelta, SKTDefaultPasteCascadeDelta);
        [[NSNotificationCenter defaultCenter] postNotificationName: SKTGraphicViewSelectionDidChangeNotification object: self];
        [self updateRulers];
    }
}

- (void) clearSelection {
    int i, c = [selectedGraphics_ count];
    id curGraphic;
    
    if( c > 0) {
        for (i=0; i<c; i++) {
            curGraphic = [selectedGraphics_ objectAtIndex: i];
            [[[self undoManager] prepareWithInvocationTarget: self] selectGraphic: curGraphic];
            [self invalidateGraphic: curGraphic];
        }
        [[self undoManager] setActionName: NSLocalizedStringFromTable(@"Selection Change", @"UndoStrings", @"Action name for selection changes.")];
        [selectedGraphics_ removeAllObjects];
        pasteCascadeDelta_ = NSMakePoint(SKTDefaultPasteCascadeDelta, SKTDefaultPasteCascadeDelta);
        [[NSNotificationCenter defaultCenter] postNotificationName: SKTGraphicViewSelectionDidChangeNotification object: self];
        [self updateRulers];
    }
}

// Editing
- (void) setEditingGraphic: (SKTGraphic*) graphic editorView: (NSView*) editorView {
    // Called by a SKTGraphic that is told to start editing.  SKTGraphicView doesn't do anything with editorView, just remembers it.
    editingGraphic_ = graphic;
    editorView_ = editorView;
}

- (SKTGraphic*) editingGraphic {
    return editingGraphic_;
}

- (NSView*) editorView {
    return editorView_;
}

- (void) startEditingGraphic: (SKTGraphic*) graphic withEvent: (NSEvent*) event {
    [graphic startEditingWithEvent: event inView: self];
}

- (void) endEditing {
    if( editingGraphic_) {
        [editingGraphic_ endEditingInView: self];
        editingGraphic_ = nil;
        editorView_ = nil;
    }
}

// Geometry calculations
- (SKTGraphic*) graphicUnderPoint: (NSPoint) point {
    SKTDrawDocument *document = [self drawDocument];
    NSArray *graphics = [document graphics];
    unsigned i, c = [graphics count];
    SKTGraphic *curGraphic = nil;

    for (i=0; i<c; i++) {
        curGraphic = [graphics objectAtIndex: i];
        if( [self mouse: point inRect: [curGraphic drawingBounds]] && [curGraphic hitTest: point isSelected: [self graphicIsSelected: curGraphic]]) {
            break;
        }
    }
    if( i < c) {
        return curGraphic;
    } else {
        return nil;
    }
}

- (NSSet*) graphicsIntersectingRect: (NSRect) rect {
    NSArray *graphics = [self graphics];
    unsigned i, c = [graphics count];
    NSMutableSet *result = [NSMutableSet set];
    SKTGraphic *curGraphic;

    for (i=0; i<c; i++) {
        curGraphic = [graphics objectAtIndex: i];
        if( NSIntersectsRect(rect, [curGraphic drawingBounds])) {
            [result addObject: curGraphic];
        }
    }
    return result;
}

// Drawing and mouse tracking
- (BOOL) isFlipped {
    return YES;
}

- (BOOL) isOpaque {
    return YES;
}

- (BOOL) acceptsFirstResponder {
    return YES;
}

- (BOOL) becomeFirstResponder {
    [self updateRulers];
    return YES;
}

- (void) drawRect: (NSRect) rect {
    SKTDrawWindowController *drawWindowController = [self drawWindowController];
    NSArray *graphics;
    unsigned i;
    SKTGraphic *curGraphic;
    BOOL isSelected;
    NSRect drawingBounds;
    NSGraphicsContext *currentContext = [NSGraphicsContext currentContext];

    [[NSColor whiteColor] set];
    NSRectFill(rect);
    if( [self showsGrid]) {
        SKTDrawGridWithSettingsInRect([self gridSpacing], [self gridColor], rect, NSZeroPoint);
    }

    graphics = [[drawWindowController document] graphics];
    i = [graphics count];
    while (i-- > 0) {
        curGraphic = [graphics objectAtIndex: i];
        drawingBounds = [curGraphic drawingBounds];
        if( NSIntersectsRect(rect, drawingBounds)) {
            if( !gvFlags_.knobsHidden && (curGraphic != editingGraphic_)) {
                // Figure out if we should draw selected.
                isSelected = [self graphicIsSelected: curGraphic];
                // Account for any current rubberband selection state
                if( rubberbandGraphics_ && (isSelected == gvFlags_.rubberbandIsDeselecting) && [rubberbandGraphics_ containsObject: curGraphic]) {
                    isSelected = (isSelected ? NO : YES);
                }
            } else {
                // Do not draw handles on graphics that are editing.
                isSelected = NO;
            }
            [currentContext saveGraphicsState];
            [NSBezierPath clipRect: drawingBounds];
            [curGraphic drawInView: self isSelected: isSelected];
            [currentContext restoreGraphicsState];
        }
    }

    if( creatingGraphic_) {
        drawingBounds = [creatingGraphic_ drawingBounds];
        if( NSIntersectsRect(rect, drawingBounds)) {
            [currentContext saveGraphicsState];
            [NSBezierPath clipRect: drawingBounds];
            [creatingGraphic_ drawInView: self isSelected: NO];
            [currentContext restoreGraphicsState];
        }
    }
    if( !NSEqualRects(rubberbandRect_, NSZeroRect)) {
        [[NSColor knobColor] set];
        NSFrameRect(rubberbandRect_);
    }
}

- (void) beginEchoingMoveToRulers: (NSRect) echoRect {
    NSRulerView *horizontalRuler = [[self enclosingScrollView] horizontalRulerView];
    NSRulerView *verticalRuler = [[self enclosingScrollView] verticalRulerView];

    horizontalRulerLineRect_ = [self convertRect: echoRect toView: horizontalRuler];
    verticalRulerLineRect_ = [self convertRect: echoRect toView: verticalRuler];

    [horizontalRuler moveRulerlineFromLocation:-1.0 toLocation: NSMinX(horizontalRulerLineRect_)];
    [horizontalRuler moveRulerlineFromLocation:-1.0 toLocation: NSMidX(horizontalRulerLineRect_)];
    [horizontalRuler moveRulerlineFromLocation:-1.0 toLocation: NSMaxX(horizontalRulerLineRect_)];

    [verticalRuler moveRulerlineFromLocation:-1.0 toLocation: NSMinY(verticalRulerLineRect_)];
    [verticalRuler moveRulerlineFromLocation:-1.0 toLocation: NSMidY(verticalRulerLineRect_)];
    [verticalRuler moveRulerlineFromLocation:-1.0 toLocation: NSMaxY(verticalRulerLineRect_)];
}

- (void) continueEchoingMoveToRulers: (NSRect) echoRect {
    NSRulerView *horizontalRuler = [[self enclosingScrollView] horizontalRulerView];
    NSRulerView *verticalRuler = [[self enclosingScrollView] verticalRulerView];
    NSRect newHorizontalRect = [self convertRect: echoRect toView: horizontalRuler];
    NSRect newVerticalRect = [self convertRect: echoRect toView: verticalRuler];

    [horizontalRuler moveRulerlineFromLocation: NSMinX(horizontalRulerLineRect_) toLocation: NSMinX(newHorizontalRect)];
    [horizontalRuler moveRulerlineFromLocation: NSMidX(horizontalRulerLineRect_) toLocation: NSMidX(newHorizontalRect)];
    [horizontalRuler moveRulerlineFromLocation: NSMaxX(horizontalRulerLineRect_) toLocation: NSMaxX(newHorizontalRect)];

    [verticalRuler moveRulerlineFromLocation: NSMinY(verticalRulerLineRect_) toLocation: NSMinY(newVerticalRect)];
    [verticalRuler moveRulerlineFromLocation: NSMidY(verticalRulerLineRect_) toLocation: NSMidY(newVerticalRect)];
    [verticalRuler moveRulerlineFromLocation: NSMaxY(verticalRulerLineRect_) toLocation: NSMaxY(newVerticalRect)];

    horizontalRulerLineRect_ = newHorizontalRect;
    verticalRulerLineRect_ = newVerticalRect;
}

- (void) stopEchoingMoveToRulers {
    NSRulerView *horizontalRuler = [[self enclosingScrollView] horizontalRulerView];
    NSRulerView *verticalRuler = [[self enclosingScrollView] verticalRulerView];

    [horizontalRuler moveRulerlineFromLocation: NSMinX(horizontalRulerLineRect_) toLocation:-1.0];
    [horizontalRuler moveRulerlineFromLocation: NSMidX(horizontalRulerLineRect_) toLocation:-1.0];
    [horizontalRuler moveRulerlineFromLocation: NSMaxX(horizontalRulerLineRect_) toLocation:-1.0];

    [verticalRuler moveRulerlineFromLocation: NSMinY(verticalRulerLineRect_) toLocation:-1.0];
    [verticalRuler moveRulerlineFromLocation: NSMidY(verticalRulerLineRect_) toLocation:-1.0];
    [verticalRuler moveRulerlineFromLocation: NSMaxY(verticalRulerLineRect_) toLocation:-1.0];

    horizontalRulerLineRect_ = NSZeroRect;
    verticalRulerLineRect_ = NSZeroRect;
}

- (void) createGraphicOfClass: (Class) theClass withEvent: (NSEvent*) theEvent {
    SKTDrawDocument *document = [self drawDocument];
    creatingGraphic_ = [[theClass allocWithZone: [document zone]] init];
    if( [creatingGraphic_ createWithEvent: theEvent inView: self]) {
        [document insertGraphic: creatingGraphic_ atIndex: 0];
        [self selectGraphic: creatingGraphic_];
        if( [creatingGraphic_ isEditable]) {
            [self startEditingGraphic: creatingGraphic_ withEvent: nil ];
        }
        [[self undoManager] setActionName: [NSString stringWithFormat: NSLocalizedStringFromTable(@"Create %@", @"UndoStrings", @"Action name for newly created graphics.  Class name is inserted at the substitution."), [[NSBundle mainBundle] localizedStringForKey: NSStringFromClass(theClass) value: @"" table: @"GraphicClassNames"]]];
		[[self recordingManager] recordCreate: creatingGraphic_];
    }
    [creatingGraphic_ release];
    creatingGraphic_ = nil;
}

- (SKTGraphic*) creatingGraphic {
    return creatingGraphic_;
}

- (void) trackKnob: (int) knob ofGraphic: (SKTGraphic*) graphic withEvent: (NSEvent*) theEvent {
    NSPoint point;
    BOOL snapsToGrid = [self snapsToGrid];
    float spacing = [self gridSpacing];
    BOOL echoToRulers = [[self enclosingScrollView] rulersVisible];

    [graphic startBoundsManipulation];
    if( echoToRulers) {
        [self beginEchoingMoveToRulers: [graphic bounds]];
    }
    while (1) {
        theEvent = [[self window] nextEventMatchingMask: (NSLeftMouseDraggedMask | NSLeftMouseUpMask)];
        point = [self convertPoint: [theEvent locationInWindow] fromView: nil];
        [self invalidateGraphic: graphic];
        if( snapsToGrid) {
            point.x = floor((point.x / spacing) + 0.5) * spacing;
            point.y = floor((point.y / spacing) + 0.5) * spacing;
        }
        knob = [graphic resizeByMovingKnob: knob toPoint: point];
        [self invalidateGraphic: graphic];
        if( echoToRulers) {
            [self continueEchoingMoveToRulers: [graphic bounds]];
        }
        if( [theEvent type] == NSLeftMouseUp) {
            break;
        }
    }
    if( echoToRulers) {
        [self stopEchoingMoveToRulers];
    }

    [graphic stopBoundsManipulation];

    [[self undoManager] setActionName: NSLocalizedStringFromTable(@"Resize", @"UndoStrings", @"Action name for resizes.")];
	[[self recordingManager] recordSize: graphic];
}

- (void) rubberbandSelectWithEvent: (NSEvent*) theEvent {
    NSPoint origPoint, curPoint;
    NSEnumerator *objEnum;
    SKTGraphic *curGraphic;

    gvFlags_.rubberbandIsDeselecting = (([theEvent modifierFlags] & NSAlternateKeyMask) ? YES : NO);
    origPoint = curPoint = [self convertPoint: [theEvent locationInWindow] fromView: nil];

    while (1) {
        theEvent = [[self window] nextEventMatchingMask: (NSLeftMouseDraggedMask | NSLeftMouseUpMask)];
        curPoint = [self convertPoint: [theEvent locationInWindow] fromView: nil];
        if( NSEqualPoints(origPoint, curPoint)) {
            if( !NSEqualRects(rubberbandRect_, NSZeroRect)) {
                [self setNeedsDisplayInRect: rubberbandRect_];
                [self performSelector: @selector(invalidateGraphic:) withEachObjectInSet: rubberbandGraphics_];
            }
            rubberbandRect_ = NSZeroRect;
            [rubberbandGraphics_ release];
            rubberbandGraphics_ = nil;
        } else {
            NSRect newRubberbandRect = SKTRectFromPoints(origPoint, curPoint);
            if( !NSEqualRects(rubberbandRect_, newRubberbandRect)) {
                [self setNeedsDisplayInRect: rubberbandRect_];
                [self performSelector: @selector(invalidateGraphic:) withEachObjectInSet: rubberbandGraphics_];
                rubberbandRect_ = newRubberbandRect;
                [rubberbandGraphics_ release];
                rubberbandGraphics_ = [[self graphicsIntersectingRect: rubberbandRect_] retain];
                [self setNeedsDisplayInRect: rubberbandRect_];
                [self performSelector: @selector(invalidateGraphic:) withEachObjectInSet: rubberbandGraphics_];
            }
        }
        if( [theEvent type] == NSLeftMouseUp) {
            break;
        }
    }

    // Now select or deselect the rubberbanded graphics.
    objEnum = [rubberbandGraphics_ objectEnumerator];
    while ((curGraphic = [objEnum nextObject]) != nil) {
        if( gvFlags_.rubberbandIsDeselecting) {
            [self deselectGraphic: curGraphic];
        } else {
            [self selectGraphic: curGraphic];
        }
    }
    if( !NSEqualRects(rubberbandRect_, NSZeroRect)) {
        [self setNeedsDisplayInRect: rubberbandRect_];
    }
   
    rubberbandRect_ = NSZeroRect;
    [rubberbandGraphics_ release];
    rubberbandGraphics_ = nil;
}

- (void) moveSelectedGraphicsWithEvent: (NSEvent*) theEvent {
    NSPoint lastPoint, curPoint;
    NSArray *selGraphics = [self selectedGraphics];
    unsigned i, c;
    SKTGraphic *graphic;
    BOOL didMove = NO, isMoving = NO;
    NSPoint selOriginOffset = NSZeroPoint;
    NSPoint boundsOrigin;
    BOOL snapsToGrid = [self snapsToGrid];
    float spacing = [self gridSpacing];
    BOOL echoToRulers = [[self enclosingScrollView] rulersVisible];
    NSRect selBounds = [[self drawDocument] boundsForGraphics: selGraphics];

    c = [selGraphics count];

    lastPoint = [self convertPoint: [theEvent locationInWindow] fromView: nil];
    if( snapsToGrid || echoToRulers) {
        selOriginOffset = NSMakePoint((lastPoint.x - selBounds.origin.x), (lastPoint.y - selBounds.origin.y));
    }
    if( echoToRulers) {
        [self beginEchoingMoveToRulers: selBounds];
    }

    while (1) {
        theEvent = [[self window] nextEventMatchingMask: (NSLeftMouseDraggedMask | NSLeftMouseUpMask)];
        curPoint = [self convertPoint: [theEvent locationInWindow] fromView: nil];
        if( !isMoving && ((fabs(curPoint.x - lastPoint.x) >= 2.0) || (fabs(curPoint.y - lastPoint.y) >= 2.0))) {
            isMoving = YES;
            [selGraphics makeObjectsPerformSelector: @selector(startBoundsManipulation)];
            gvFlags_.knobsHidden = YES;
        }
        if( isMoving) {
            if( snapsToGrid) {
                boundsOrigin.x = curPoint.x - selOriginOffset.x;
                boundsOrigin.y = curPoint.y - selOriginOffset.y;
                boundsOrigin.x = floor((boundsOrigin.x / spacing) + 0.5) * spacing;
                boundsOrigin.y = floor((boundsOrigin.y / spacing) + 0.5) * spacing;
                curPoint.x = boundsOrigin.x + selOriginOffset.x;
                curPoint.y = boundsOrigin.y + selOriginOffset.y;
            }
            if( !NSEqualPoints(lastPoint, curPoint)) {
                for (i=0; i<c; i++) {
                    graphic = [selGraphics objectAtIndex: i];
                    [self invalidateGraphic: graphic];
                    [graphic moveBy: NSMakePoint(curPoint.x - lastPoint.x, curPoint.y - lastPoint.y)];
                    [self invalidateGraphic: graphic];
                    if( echoToRulers) {
                        [self continueEchoingMoveToRulers: NSMakeRect(curPoint.x - selOriginOffset.x, curPoint.y - selOriginOffset.y, NSWidth(selBounds),NSHeight(selBounds))];
                    }
                    didMove = YES;
                }
                // Adjust the delta that is used for cascading pastes.  Pasting and then moving the pasted graphic is the way you determine the cascade delta for subsequent pastes.
                pasteCascadeDelta_.x += (curPoint.x - lastPoint.x);
                pasteCascadeDelta_.y += (curPoint.y - lastPoint.y);
            }
            lastPoint = curPoint;
        }
        if( [theEvent type] == NSLeftMouseUp) {
            break;
        }
    }

    if( echoToRulers)  {
        [self stopEchoingMoveToRulers];
    }
    if( isMoving) {
        [selGraphics makeObjectsPerformSelector: @selector(stopBoundsManipulation)];
        gvFlags_.knobsHidden = NO;

        if( didMove) {
            // Only if we really moved.
            [[self undoManager] setActionName: NSLocalizedStringFromTable(@"Move", @"UndoStrings", @"Action name for moves.")];
			[[self recordingManager] recordPosition: selGraphics];
        }
    }
}

- (void) selectAndTrackMouseWithEvent: (NSEvent*) theEvent {
    NSPoint curPoint;
    SKTGraphic *graphic = nil;
    BOOL isSelected;
    BOOL extending = (([theEvent modifierFlags] & NSShiftKeyMask) ? YES : NO);

    curPoint = [self convertPoint: [theEvent locationInWindow] fromView: nil];
    graphic = [self graphicUnderPoint: curPoint];
    isSelected = (graphic ? [self graphicIsSelected: graphic] : NO);

    if( !extending && !isSelected) {
        [self clearSelection];
    }

    if( graphic) {
        // Add or remove this graphic from selection.
        if( extending) {
            if( isSelected) {
                [self deselectGraphic: graphic];
                isSelected = NO;
            } else {
                [self selectGraphic: graphic];
                isSelected = YES;
            }
        } else {
            if( isSelected) {
                int knobHit = [graphic knobUnderPoint: curPoint];
                if( knobHit != NoKnob) {
                    [self trackKnob: knobHit ofGraphic: graphic withEvent: theEvent];
                    return;
                }
            }
            [self selectGraphic: graphic];
            isSelected = YES;
        }
    } else {
        [self rubberbandSelectWithEvent: theEvent];
        return;
    }

    if( isSelected) {
        [self moveSelectedGraphicsWithEvent: theEvent];
        return;
    }

    // If we got here then there must be nothing else to do.  Just track until mouseUp:.
    while (1) {
        theEvent = [[self window] nextEventMatchingMask: (NSLeftMouseDraggedMask | NSLeftMouseUpMask)];
        if( [theEvent type] == NSLeftMouseUp) {
            break;
        }
    }
}

- (void) mouseDown: (NSEvent*) theEvent {
    Class theClass = [[SKTToolPaletteController sharedToolPaletteController] currentGraphicClass];
    if( [self editingGraphic]) {
        [self endEditing];
    }
    if( [theEvent clickCount] > 1) {
        NSPoint curPoint = [self convertPoint: [theEvent locationInWindow] fromView: nil];
        SKTGraphic *graphic = [self graphicUnderPoint: curPoint];
        if( graphic && [graphic isEditable]) {
            [self startEditingGraphic: graphic withEvent: theEvent];
            return;
        }
    }
    if( theClass) {
        [self clearSelection];
        [self createGraphicOfClass: theClass withEvent: theEvent];
    } else {
        [self selectAndTrackMouseWithEvent: theEvent];
    }
}

- (BOOL) acceptsFirstMouse: (NSEvent*) theEvent {
    return YES;
}

// SKTImage graphic creation
- (BOOL) makeNewImageFromPasteboard: (NSPasteboard*) pboard atPoint: (NSPoint) point {
    NSString *type = [pboard availableTypeFromArray: [NSImage imagePasteboardTypes]];
    if( type) {
        NSImage *contents = [[NSImage allocWithZone: [[self drawDocument] zone]] initWithPasteboard: pboard];
        if( contents) {
            SKTImage *newImage = [[SKTImage allocWithZone: [[self drawDocument] zone]] init];
            [newImage setBounds: NSMakeRect(point.x, point.y - [contents size].height, [contents size].width, [contents size].height)];
            [newImage setImage: contents];
            [contents release];
            [[self drawDocument] insertGraphic: newImage atIndex: 0];
            [newImage release];
            [self clearSelection];
            [self selectGraphic: newImage];
			[[self recordingManager] recordCreate: newImage];
            return YES;
        }
    }
    return NO;
}

- (BOOL) makeNewImageFromContentsOfFile: (NSString*) filename atPoint: (NSPoint) point {
    NSString *extension = [filename pathExtension];
    if( [[NSImage imageFileTypes] containsObject: extension]) {
        NSImage *contents = [[NSImage allocWithZone: [[self drawDocument] zone]] initWithContentsOfFile: filename];
        if( contents) {
            SKTImage *newImage = [[SKTImage allocWithZone: [[self drawDocument] zone]] init];
            [newImage setBounds: NSMakeRect(point.x, point.y, [contents size].width, [contents size].height)];
            [newImage setImage: contents];
            [contents release];
            [[self drawDocument] insertGraphic: newImage atIndex: 0];
            [newImage release];
            [self clearSelection];
            [self selectGraphic: newImage];
			[[self recordingManager] recordCreate: newImage];
            return YES;
        }
    }
    return NO;
}

// Dragging
- (unsigned int) dragOperationForDraggingInfo: (id <NSDraggingInfo>) sender {
    NSPasteboard *pboard = [sender draggingPasteboard];
    NSString *type = [pboard availableTypeFromArray: [NSArray arrayWithObjects: NSColorPboardType, NSFilenamesPboardType, nil]];
    
    if( type) {
        if( [type isEqualToString: NSColorPboardType]) {
            NSPoint point = [self convertPoint: [sender draggingLocation] fromView: nil];
            if( [self graphicUnderPoint: point]) {
                return NSDragOperationGeneric;
            }
        }
        if( [type isEqualToString: NSFilenamesPboardType]) {
            return NSDragOperationCopy;
        }
    }

    type = [pboard availableTypeFromArray: [NSImage imagePasteboardTypes]];
    if( type) {
        return NSDragOperationCopy;
    }
    
    return NSDragOperationNone;
}

- (unsigned int) draggingEntered: (id <NSDraggingInfo>) sender {
    return [self dragOperationForDraggingInfo: sender];
}

- (unsigned int) draggingUpdated: (id <NSDraggingInfo>) sender {
    return [self dragOperationForDraggingInfo: sender];
}

- (void) draggingExited: (id <NSDraggingInfo>) sender {
    return;
}

- (BOOL) prepareForDragOperation: (id <NSDraggingInfo>) sender {
    return YES;
}

- (BOOL) performDragOperation: (id <NSDraggingInfo>) sender {
    return YES;
}

- (void) concludeDragOperation: (id <NSDraggingInfo>) sender {
    NSPasteboard *pboard = [sender draggingPasteboard];
    NSString *type = [pboard availableTypeFromArray: [NSArray arrayWithObjects: NSColorPboardType, NSFilenamesPboardType, nil]];
    NSPoint point = [self convertPoint: [sender draggingLocation] fromView: nil];
    NSPoint draggedImageLocation = [self convertPoint: [sender draggedImageLocation] fromView: nil];

    if( type) {
        if( [type isEqualToString: NSColorPboardType]) {
            SKTGraphic *hitGraphic = [self graphicUnderPoint: point];
            
            if( hitGraphic) {
                NSColor *color = [[NSColor colorFromPasteboard: pboard] colorWithAlphaComponent: 1.0];
                [hitGraphic setFillColor: color];
                [[self undoManager] setActionName: NSLocalizedStringFromTable(@"Set Fill Color", @"UndoStrings", @"Action name for setting fill color.")];
				[[self recordingManager] recordSet: hitGraphic property: @"fillColor" to: color];
            }
        } else if( [type isEqualToString: NSFilenamesPboardType]) {
            NSArray *filenames = [pboard propertyListForType: NSFilenamesPboardType];
            // Handle multiple files (cascade them?)
            if( [filenames count] == 1) {
                NSString *filename = [filenames objectAtIndex: 0];
                [self makeNewImageFromContentsOfFile: filename atPoint: point];
            }
        }
        return;
    }

    (void)[self makeNewImageFromPasteboard: pboard atPoint: draggedImageLocation];
}

// Ruler support
- (void) updateRulers {
    NSScrollView *enclosingScrollView = [self enclosingScrollView];
    if( [enclosingScrollView rulersVisible]) {
        // MF: Eventually, it'd be nice if we added ruler markers for the selection, but for now we just clear them.  By clearing the markers we make sure that no markers from text editing are left over when the editing stops.
        [[enclosingScrollView verticalRulerView] setMarkers: nil];
        [[enclosingScrollView horizontalRulerView] setMarkers: nil];
    }
}

- (BOOL) rulerView: (NSRulerView*) ruler shouldMoveMarker: (NSRulerMarker*) marker {
    return YES;
}

- (float) rulerView: (NSRulerView*) ruler willMoveMarker: (NSRulerMarker*) marker toLocation: (float) location {
    return location;
}

- (void) rulerView: (NSRulerView*) ruler didMoveMarker: (NSRulerMarker*) marker {
    
}

- (BOOL) rulerView: (NSRulerView*) ruler shouldRemoveMarker: (NSRulerMarker*) marker {
    return NO;
}

#define SKT_RULER_MARKER_THICKNESS 8.0
#define SKT_RULER_ACCESSORY_THICKNESS 10.0

- (IBAction) toggleRuler: (id) sender {
    NSScrollView *enclosingScrollView = [self enclosingScrollView];
    BOOL rulersAreVisible = [enclosingScrollView rulersVisible];
    if( rulersAreVisible) {
        [enclosingScrollView setRulersVisible: NO];
    } else {
        if( !gvFlags_.initedRulers) {
            NSRulerView *ruler;
            ruler = [enclosingScrollView horizontalRulerView];
            [ruler setReservedThicknessForMarkers: SKT_RULER_MARKER_THICKNESS];
            [ruler setReservedThicknessForAccessoryView: SKT_RULER_ACCESSORY_THICKNESS];
            ruler = [enclosingScrollView verticalRulerView];
            [ruler setReservedThicknessForMarkers: SKT_RULER_MARKER_THICKNESS];
            [ruler setReservedThicknessForAccessoryView: SKT_RULER_ACCESSORY_THICKNESS];
            gvFlags_.initedRulers = YES;
        }
        [enclosingScrollView setRulersVisible: YES];
        [self updateRulers];
    }
}

// Action methods and other UI entry points
- (void) changeColor: (id) sender {
    NSArray *selGraphics = [self selectedGraphics];
    unsigned i, c = [selGraphics count];
    if( c > 0) {
        SKTGraphic *curGraphic;
        NSColor *color = [sender color];

        for (i=0; i<c; i++) {
            curGraphic = [selGraphics objectAtIndex: i];
            [curGraphic setFillColor: color];
            [curGraphic setDrawsFill: YES];
        }
        [[self undoManager] setActionName: NSLocalizedStringFromTable(@"Set Fill Color", @"UndoStrings", @"Action name for setting fill color.")];
		[[self recordingManager] recordSet: selGraphics property: @"fillColor" to: color];
    }
}

- (IBAction) selectAll: (id) sender {
    NSArray *graphics = [[self drawDocument] graphics];
    [self performSelector: @selector(selectGraphic:) withEachObjectInArray: graphics];
}

- (IBAction) deselectAll: (id) sender {
    [self clearSelection];
}

- (IBAction) delete: (id) sender {
    NSArray *selCopy = [[NSArray allocWithZone: [self zone]] initWithArray: [self selectedGraphics]];
    if( [selCopy count] > 0) {
		[[self recordingManager] recordDelete: selCopy];
        [[self drawDocument] performSelector: @selector(removeGraphic:) withEachObjectInArray: selCopy];
        [selCopy release];
        [[self undoManager] setActionName: NSLocalizedStringFromTable(@"Delete", @"UndoStrings", @"Action name for deletions.")];
    }
}

- (IBAction) bringToFront: (id) sender {
    NSArray *orderedSelection = [self orderedSelectedGraphics];
    unsigned c = [orderedSelection count];
    if( c > 0) {
        SKTDrawDocument *document = [self drawDocument];
        while (c-- > 0) {
            [document moveGraphic: [orderedSelection objectAtIndex: c] toIndex: 0];
        }
        [[self undoManager] setActionName: NSLocalizedStringFromTable(@"Bring To Front", @"UndoStrings", @"Action name for bring to front.")];
		[[self recordingManager] recordIndex: orderedSelection];
    }
}

- (IBAction) sendToBack: (id) sender {
    NSArray *orderedSelection = [self orderedSelectedGraphics];
    unsigned i, c = [orderedSelection count];
    if( c > 0) {
        SKTDrawDocument *document = [self drawDocument];
        unsigned lastIndex = [[self graphics] count];
        for (i=0; i<c; i++) {
            [document moveGraphic: [orderedSelection objectAtIndex: i] toIndex: lastIndex];
        }
        [[self undoManager] setActionName: NSLocalizedStringFromTable(@"Send To Back", @"UndoStrings", @"Action name for send to back.")];
		[[self recordingManager] recordIndex: orderedSelection];
    }
}

- (IBAction) alignLeftEdges: (id) sender {
    NSArray *selection = [self selectedGraphics];
    unsigned i, c = [selection count];
    if( c > 1) {
        NSRect firstBounds = [[selection objectAtIndex: 0] bounds];
        SKTGraphic *curGraphic;
        NSRect curBounds;
        for (i=1; i<c; i++) {
            curGraphic = [selection objectAtIndex: i];
            curBounds = [curGraphic bounds];
            if( curBounds.origin.x != firstBounds.origin.x) {
                curBounds.origin.x = firstBounds.origin.x;
                [curGraphic setBounds: curBounds];
            }
        }
        [[self undoManager] setActionName: NSLocalizedStringFromTable(@"Align Left Edges", @"UndoStrings", @"Action name for align left edges.")];
		[[self recordingManager] recordPosition: selection];
    }
}

- (IBAction) alignRightEdges: (id) sender {
    NSArray *selection = [self selectedGraphics];
    unsigned i, c = [selection count];
    if( c > 1) {
        NSRect firstBounds = [[selection objectAtIndex: 0] bounds];
        SKTGraphic *curGraphic;
        NSRect curBounds;
        for (i=1; i<c; i++) {
            curGraphic = [selection objectAtIndex: i];
            curBounds = [curGraphic bounds];
            if( NSMaxX(curBounds) != NSMaxX(firstBounds)) {
                curBounds.origin.x = NSMaxX(firstBounds) - curBounds.size.width;
                [curGraphic setBounds: curBounds];
            }
        }
        [[self undoManager] setActionName: NSLocalizedStringFromTable(@"Align Right Edges", @"UndoStrings", @"Action name for align right edges.")];
		[[self recordingManager] recordPosition: selection];
    }
}

- (IBAction) alignTopEdges: (id) sender {
    NSArray *selection = [self selectedGraphics];
    unsigned i, c = [selection count];
    if( c > 1) {
        NSRect firstBounds = [[selection objectAtIndex: 0] bounds];
        SKTGraphic *curGraphic;
        NSRect curBounds;
        for (i=1; i<c; i++) {
            curGraphic = [selection objectAtIndex: i];
            curBounds = [curGraphic bounds];
            if( curBounds.origin.y != firstBounds.origin.y) {
                curBounds.origin.y = firstBounds.origin.y;
                [curGraphic setBounds: curBounds];
            }
        }
        [[self undoManager] setActionName: NSLocalizedStringFromTable(@"Align Top Edges", @"UndoStrings", @"Action name for align top edges.")];
		[[self recordingManager] recordPosition: selection];
    }
}

- (IBAction) alignBottomEdges: (id) sender {
    NSArray *selection = [self selectedGraphics];
    unsigned i, c = [selection count];
    if( c > 1) {
        NSRect firstBounds = [[selection objectAtIndex: 0] bounds];
        SKTGraphic *curGraphic;
        NSRect curBounds;
        for (i=1; i<c; i++) {
            curGraphic = [selection objectAtIndex: i];
            curBounds = [curGraphic bounds];
            if( NSMaxY(curBounds) != NSMaxY(firstBounds)) {
                curBounds.origin.y = NSMaxY(firstBounds) - curBounds.size.height;
                [curGraphic setBounds: curBounds];
            }
        }
        [[self undoManager] setActionName: NSLocalizedStringFromTable(@"Align Bottom Edges", @"UndoStrings", @"Action name for align bottom edges.")];
		[[self recordingManager] recordPosition: selection];
    }
}

- (IBAction) alignHorizontalCenters: (id) sender {
    NSArray *selection = [self selectedGraphics];
    unsigned i, c = [selection count];
    if( c > 1) {
        NSRect firstBounds = [[selection objectAtIndex: 0] bounds];
        SKTGraphic *curGraphic;
        NSRect curBounds;
        for (i=1; i<c; i++) {
            curGraphic = [selection objectAtIndex: i];
            curBounds = [curGraphic bounds];
            if( NSMidX(curBounds) != NSMidX(firstBounds)) {
                curBounds.origin.x = NSMidX(firstBounds) - (curBounds.size.width / 2.0);
                [curGraphic setBounds: curBounds];
            }
        }
        [[self undoManager] setActionName: NSLocalizedStringFromTable(@"Align Horizontal Centers", @"UndoStrings", @"Action name for align horizontal centers.")];
		[[self recordingManager] recordPosition: selection];
    }
}

- (IBAction) alignVerticalCenters: (id) sender {
    NSArray *selection = [self selectedGraphics];
    unsigned i, c = [selection count];
    if( c > 1) {
        NSRect firstBounds = [[selection objectAtIndex: 0] bounds];
        SKTGraphic *curGraphic;
        NSRect curBounds;
        for (i=1; i<c; i++) {
            curGraphic = [selection objectAtIndex: i];
            curBounds = [curGraphic bounds];
            if( NSMidY(curBounds) != NSMidY(firstBounds)) {
                curBounds.origin.y = NSMidY(firstBounds) - (curBounds.size.height / 2.0);
                [curGraphic setBounds: curBounds];
            }
        }
        [[self undoManager] setActionName: NSLocalizedStringFromTable(@"Align Vertical Centers", @"UndoStrings", @"Action name for align vertical centers.")];
		[[self recordingManager] recordPosition: selection];
    }
}

- (IBAction) makeSameWidth: (id) sender {
    NSArray *selection = [self selectedGraphics];
    unsigned i, c = [selection count];
    if( c > 1) {
        NSRect firstBounds = [[selection objectAtIndex: 0] bounds];
        SKTGraphic *curGraphic;
        NSRect curBounds;
        for (i=1; i<c; i++) {
            curGraphic = [selection objectAtIndex: i];
            curBounds = [curGraphic bounds];
            if( curBounds.size.width != firstBounds.size.width) {
                curBounds.size.width = firstBounds.size.width;
                [curGraphic setBounds: curBounds];
            }
        }
        [[self undoManager] setActionName: NSLocalizedStringFromTable(@"Make Same Width", @"UndoStrings", @"Action name for make same width.")];
		[[self recordingManager] recordPosition: selection];
    }
}

- (IBAction) makeSameHeight: (id) sender {
    NSArray *selection = [self selectedGraphics];
    unsigned i, c = [selection count];
    if( c > 1) {
        NSRect firstBounds = [[selection objectAtIndex: 0] bounds];
        SKTGraphic *curGraphic;
        NSRect curBounds;
        for (i=1; i<c; i++) {
            curGraphic = [selection objectAtIndex: i];
            curBounds = [curGraphic bounds];
            if( curBounds.size.height != firstBounds.size.height) {
                curBounds.size.height = firstBounds.size.height;
                [curGraphic setBounds: curBounds];
            }
        }
        [[self undoManager] setActionName: NSLocalizedStringFromTable(@"Make Same Width", @"UndoStrings", @"Action name for make same width.")];
		[[self recordingManager] recordSize: selection];
    }
}

- (IBAction) makeNaturalSize: (id) sender {
    NSArray *selection = [self selectedGraphics];
    if( [selection count] > 0) {
        [selection makeObjectsPerformSelector: @selector(makeNaturalSize)];
        [[self undoManager] setActionName: NSLocalizedStringFromTable(@"Make Natural Size", @"UndoStrings", @"Action name for natural size.")];
		[[self recordingManager] recordSize: selection];
    }
}

- (IBAction) snapsToGridMenuAction: (id) sender {
    [self setSnapsToGrid: ([sender state] ? NO : YES)];
    // Menu item will get state fixed up in validateMenuItem:
    [[SKTGridPanelController sharedGridPanelController] updatePanel];
}

- (IBAction) showsGridMenuAction: (id) sender {
    [self setShowsGrid: ([sender state] ? NO : YES)];
    // Menu item will get state fixed up in validateMenuItem:
    [[SKTGridPanelController sharedGridPanelController] updatePanel];
}

- (IBAction) gridSelectedGraphicsAction: (id) sender {
    NSArray *selection = [self selectedGraphics];
    unsigned i, c = [selection count];
    if( c > 0) {
        SKTGraphic *curGraphic;
        NSRect curBounds;
        NSPoint curMaxPoint;
        float spacing = [self gridSpacing];
        
        for (i=0; i<c; i++) {
            curGraphic = [selection objectAtIndex: i];
            curBounds = [curGraphic bounds];
            curMaxPoint = NSMakePoint(NSMaxX(curBounds), NSMaxY(curBounds));
            curBounds.origin.x = floor((curBounds.origin.x / spacing) + 0.5) * spacing;
            curBounds.origin.y = floor((curBounds.origin.y / spacing) + 0.5) * spacing;
            curMaxPoint.x = floor((curMaxPoint.x / spacing) + 0.5) * spacing;
            curMaxPoint.y = floor((curMaxPoint.y / spacing) + 0.5) * spacing;
            curBounds.size.width = curMaxPoint.x - curBounds.origin.x;
            curBounds.size.height = curMaxPoint.y - curBounds.origin.y;
            [curGraphic setBounds: curBounds];
        }
        [[self undoManager] setActionName: NSLocalizedStringFromTable(@"Grid Selected Graphics", @"UndoStrings", @"Action name for grid selected graphics.")];
		[[self recordingManager] recordBounds: selection];
    }
}

- (BOOL) validateMenuItem: (NSMenuItem*) item {
    SEL action = [item action];

    if( action == @selector(snapsToGridMenuAction:)) {
        [item setState: ([self snapsToGrid] ? NSOnState : NSOffState)];
        return YES;
    } else if( action == @selector(showsGridMenuAction:)) {
        [item setState: ([self showsGrid] ? NSOnState : NSOffState)];
        return YES;
    } else if( action == @selector(makeNaturalSize:)) {
        // Return YES if we have at least one selected graphic that has a natural size.
        NSArray *selectedGraphics = [self selectedGraphics];
        unsigned i, c = [selectedGraphics count];
        if( c > 0) {
            for (i=0; i<c; i++) {
                if( [[selectedGraphics objectAtIndex: i] hasNaturalSize]) {
                    return YES;
                }
            }
        }
        return NO;
    } else if( (action == @selector(gridSelectedGraphicsAction:)) || (action == @selector(delete:)) || (action == @selector(bringToFront:)) || (action == @selector(sendToBack:)) || (action == @selector(cut:)) || (action == @selector(copy:))) {
        // These only apply if there is a selection
        return (([[self selectedGraphics] count] > 0) ? YES : NO);
    } else if( (action == @selector(alignLeftEdges:)) || (action == @selector(alignRightEdges:)) || (action == @selector(alignTopEdges:)) || (action == @selector(alignBottomEdges:)) || (action == @selector(alignHorizontalCenters:)) || (action == @selector(alignVerticalCenters:)) || (action == @selector(alignTextBaselines:)) || (action == @selector(makeSameWidth:)) || (action == @selector(makeSameHeight:))) {
        // These only apply to multiple selection
        return (([[self selectedGraphics] count] > 1) ? YES : NO);
    } else {
        return YES;
    }
}

- (IBAction) copy: (id) sender {
    NSArray *orderedSelection = [self orderedSelectedGraphics];
    if( [orderedSelection count] > 0) {
        SKTDrawDocument *document = [self drawDocument];
        NSPasteboard *pboard = [NSPasteboard generalPasteboard];

        [pboard declareTypes: [NSArray arrayWithObjects: SKTDrawDocumentType, NSTIFFPboardType, NSPDFPboardType, nil] owner: nil];
        [pboard setData: [document drawDocumentDataForGraphics: orderedSelection] forType: SKTDrawDocumentType];
        [pboard setData: [document TIFFRepresentationForGraphics: orderedSelection error: NULL] forType: NSTIFFPboardType];
        [pboard setData: [document PDFRepresentationForGraphics: orderedSelection] forType: NSPDFPboardType];
        pasteboardChangeCount_ = [pboard changeCount];
        pasteCascadeNumber_ = 1;
        pasteCascadeDelta_ = NSMakePoint(SKTDefaultPasteCascadeDelta, SKTDefaultPasteCascadeDelta);
    }
}

- (IBAction) cut: (id) sender {
    [self copy: sender];
    [self delete: sender];
    [[self undoManager] setActionName: NSLocalizedStringFromTable(@"Cut", @"UndoStrings", @"Action name for cut.")];
}

- (IBAction) paste: (id) sender {
    NSPasteboard *pboard = [NSPasteboard generalPasteboard];
    NSString *type = [pboard availableTypeFromArray: [NSArray arrayWithObjects: SKTDrawDocumentType, NSFilenamesPboardType, nil]];

    if( type) {
        if( [type isEqualToString: SKTDrawDocumentType]) {

            // You can't trust anything that might have been put on the pasteboard by another application, so be ready for -[SKTDrawDocument drawDocumentDictionaryFromData: error:] to fail and return nil.
            NSError *error;
            SKTDrawDocument *document = [self drawDocument];
            NSDictionary *docDict = [document drawDocumentDictionaryFromData: [pboard dataForType: type] error:&error];
            if( docDict) {
                NSArray *array = [document graphicsFromDrawDocumentDictionary: docDict];
                int i = [array count];
                int currentChangeCount = [pboard changeCount];
                
                if( pasteboardChangeCount_ != currentChangeCount) {
                    pasteboardChangeCount_ = currentChangeCount;
                    pasteCascadeNumber_ = 0;
                    pasteCascadeDelta_ = NSMakePoint(SKTDefaultPasteCascadeDelta, SKTDefaultPasteCascadeDelta);
                }

                if( i > 0) {
                    id curGraphic;
                    NSPoint savedPasteCascadeDelta = pasteCascadeDelta_;
                    
                    [self clearSelection];
                    while (i-- > 0) {
                        curGraphic = [array objectAtIndex: i];
                        if( pasteCascadeNumber_ > 0) {
                            [curGraphic moveBy: NSMakePoint(pasteCascadeNumber_ * savedPasteCascadeDelta.x, pasteCascadeNumber_ * savedPasteCascadeDelta.y)];
                        }
                        [document insertGraphic: curGraphic atIndex: 0];
                        [self selectGraphic: curGraphic];
                    }
                    pasteCascadeNumber_++;
                    pasteCascadeDelta_ = savedPasteCascadeDelta;
                    [[self undoManager] setActionName: NSLocalizedStringFromTable(@"Paste", @"UndoStrings", @"Action name for paste.")];
                }
            } else {

                // Something went wrong? Present the error to the user in a sheet. It was entirely -[NSDocument drawDocumentDictionaryFromData: error:]'s responsibility to set error to something when it returned nil. It was also entirely responsible for not crashing if we had passed in error: NULL.
                [self presentError: error modalForWindow: [self window] delegate: nil didPresentSelector: NULL contextInfo: NULL];

            }

        } else if( [type isEqualToString: NSFilenamesPboardType]) {
            NSArray *filenames = [pboard propertyListForType: NSFilenamesPboardType];
            if( [filenames count] == 1) {
                NSString *filename = [filenames objectAtIndex: 0];
                if( [self makeNewImageFromContentsOfFile: filename atPoint: NSMakePoint(50, 50)]) {
                    [[self undoManager] setActionName: NSLocalizedStringFromTable(@"Paste", @"UndoStrings", @"Action name for paste.")];
                }
            }
        }
    } else if( [self makeNewImageFromPasteboard: pboard atPoint: NSMakePoint(50, 50)]) {
        [[self undoManager] setActionName: NSLocalizedStringFromTable(@"Paste", @"UndoStrings", @"Action name for paste.")];
    }
}

// Keyboard commands
- (void) keyDown: (NSEvent*) event {
    // Pass on the key binding manager.  This will end up calling insertText: or some command selector.
    [self interpretKeyEvents: [NSArray arrayWithObject: event]];
}

- (void) insertText: (NSString*) str {
    NSBeep();
}

- (void) hideKnobsMomentarily {
    if( unhideKnobsTimer_) {
        [unhideKnobsTimer_ invalidate];
        unhideKnobsTimer_ = nil;
    }
    unhideKnobsTimer_ = [NSTimer scheduledTimerWithTimeInterval: 0.5 target: self selector: @selector(unhideKnobs:) userInfo: nil repeats: NO];
    gvFlags_.knobsHidden = YES;
    [self invalidateGraphics: [self selectedGraphics]];
}

- (void) unhideKnobs: (NSTimer*) timer {
    gvFlags_.knobsHidden = NO;
    [self invalidateGraphics: [self selectedGraphics]];
    [unhideKnobsTimer_ invalidate];
    unhideKnobsTimer_ = nil;
}

- (void) moveSelectedGraphicsByPoint: (NSPoint) delta {
    NSArray *selection = [self selectedGraphics];
    unsigned i, c = [selection count];
    if( c > 0) {
        [self hideKnobsMomentarily];
        for (i=0; i<c; i++) {
            [[selection objectAtIndex: i] moveBy: delta];
        }
        [[self undoManager] setActionName: NSLocalizedStringFromTable(@"Nudge", @"UndoStrings", @"Action name for nudge keyboard commands.")];
		[[self recordingManager] recordPosition: selection];
    }
}

- (void) moveLeft: (id) sender {
    [self moveSelectedGraphicsByPoint: NSMakePoint(-1.0, 0.0)];
}

- (void) moveRight: (id) sender {
    [self moveSelectedGraphicsByPoint: NSMakePoint(1.0, 0.0)];
}

- (void) moveUp: (id) sender {
    [self moveSelectedGraphicsByPoint: NSMakePoint(0.0, -1.0)];
}

- (void) moveDown: (id) sender {
    [self moveSelectedGraphicsByPoint: NSMakePoint(0.0, 1.0)];
}

- (void) moveForwardAndModifySelection: (id) sender {
    // We will use this to move by the grid spacing.
    [self moveSelectedGraphicsByPoint: NSMakePoint([self gridSpacing], 0.0)];
}

- (void) moveBackwardAndModifySelection: (id) sender {
    // We will use this to move by the grid spacing.
    [self moveSelectedGraphicsByPoint: NSMakePoint(-[self gridSpacing], 0.0)];
}

- (void) moveUpAndModifySelection: (id) sender {
    // We will use this to move by the grid spacing.
    [self moveSelectedGraphicsByPoint: NSMakePoint(0.0, -[self gridSpacing])];
}

- (void) moveDownAndModifySelection: (id) sender {
    // We will use this to move by the grid spacing.
    [self moveSelectedGraphicsByPoint: NSMakePoint(0.0, [self gridSpacing])];
}

- (void) deleteForward: (id) sender {
    [self delete: sender];
}

- (void) deleteBackward: (id) sender {
    [self delete: sender];
}

// Grid settings
- (BOOL) snapsToGrid {
    return gvFlags_.snapsToGrid;
}

- (void) setSnapsToGrid: (BOOL) flag {
    gvFlags_.snapsToGrid = flag;
}

- (BOOL) showsGrid {
    return gvFlags_.showsGrid;
}

- (void) setShowsGrid: (BOOL) flag {
    if( gvFlags_.showsGrid != flag) {
        gvFlags_.showsGrid = flag;
        [self setNeedsDisplay: YES];
    }
}

- (float) gridSpacing {
    return gridSpacing_;
}

- (void) setGridSpacing: (float) spacing {
    if( gridSpacing_ != spacing) {
        gridSpacing_ = spacing;
        [self setNeedsDisplay: YES];
    }
}

- (NSColor*) gridColor {
    return (gridColor_ ? gridColor_ : [NSColor lightGrayColor]);
}

- (void) setGridColor: (NSColor*) color {
    if( gridColor_ != color) {
        [gridColor_ release];
        gridColor_ = [color retain];
        [self setNeedsDisplay: YES];
    }
}

- (SFRecordingManager*) recordingManager {
	return [SFRecordingManager sharedInstance];
}

@end


/*
 IMPORTANT:  This Apple software is supplied to you by Apple Computer, Inc. ("Apple") in
 consideration of your agreement to the following terms, and your use, installation,
 modification or redistribution of this Apple software constitutes acceptance of these
 terms.  If you do not agree with these terms, please do not use, install, modify or
 redistribute this Apple software.

 In consideration of your agreement to abide by the following terms, and subject to these
 terms, Apple grants you a personal, non-exclusive license, under Apple's copyrights in
 this original Apple software (the "Apple Software"), to use, reproduce, modify and
 redistribute the Apple Software, with or without modifications, in source and/or binary
 forms; provided that if you redistribute the Apple Software in its entirety and without
 modifications, you must retain this notice and the following text and disclaimers in all
 such redistributions of the Apple Software.  Neither the name, trademarks, service marks
 or logos of Apple Computer, Inc. may be used to endorse or promote products derived from
 the Apple Software without specific prior written permission from Apple. Except as expressly
 stated in this notice, no other rights or licenses, express or implied, are granted by Apple
 herein, including but not limited to any patent rights that may be infringed by your
 derivative works or by other works in which the Apple Software may be incorporated.

 The Apple Software is provided by Apple on an "AS IS" basis.  APPLE MAKES NO WARRANTIES,
 EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF NON-INFRINGEMENT,
 MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, REGARDING THE APPLE SOFTWARE OR ITS
 USE AND OPERATION ALONE OR IN COMBINATION WITH YOUR PRODUCTS.

 IN NO EVENT SHALL APPLE BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
          OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) ARISING IN ANY WAY OUT OF THE USE,
 REPRODUCTION, MODIFICATION AND/OR DISTRIBUTION OF THE APPLE SOFTWARE, HOWEVER CAUSED AND
 WHETHER UNDER THEORY OF CONTRACT, TORT (INCLUDING NEGLIGENCE), STRICT LIABILITY OR
 OTHERWISE, EVEN IF APPLE HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// Some modifications by David Phillip Oster. You may use those modifications freely, provided that just
// as you hold Apple blameless, you must also hold me blameless. Feb 2006 <mailto:oster@ieee.org>
 */
